/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file canvasInteract.cpp
@ingroup ExamplePrograms
@brief Example application showing how to interact with an QVImageCanvas object.

This example lets the user to play with the ROI selector of the QVImageCanvas widget, marking the selected region in the output image with a dark square in the image canvas.

You must specify a valid input video source to execute the application. Then, clicking on the ROI selector button:

@image html qvimagecanvas_roi.png

You can select a ROI that will be painted darker in the output image by the application.

@section Usage Usage of the program.
Once installed the QVision, compile using the <i>qmake</i> and <i>make</i> tools:
@code # qmake; make @endcode
Execute the application with the following line:
@code ./roroscoper --URL=<source video identifier string> @endcode
For example:
@code ./roroscoper --URL=http://perception.inf.um.es/public_data/videos/misc/minuto.avi @endcode
See section @ref QVMPlayerURLFormat for further info about the format and types of source video identifier strings.

@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <QVApplication>
#include <QVMPlayerCamera>
#include <QVDefaultGUI>
#include <QVImageCanvas>
#include <qvipp.h>

#define RADIUS 1

#ifndef DOXYGEN_IGNORE_THIS
class MyWorker: public QVWorker
	{
	public:
		MyWorker(QString name): QVWorker(name)
			{
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,1> >("Output image", outputFlag);
			addProperty<QRect>("in rect selec", inputFlag);
			addProperty<QVPolyline>("in polyline", inputFlag);
			}

		void iterate()
			{
			QVImage<uChar> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			QRect rect = getPropertyValue<QRect>("in rect selec");
			QVPolyline poly = getPropertyValue<QVPolyline>("in polyline");
			QVImage<uChar> dest = image;

			// get dark in the ROI area
			if (rect != QRect()) image.setROI(rect);
			AddC(image, 10, dest, 1, QPoint(rect.x(), rect.y()));
			dest.resetROI();

			// undo it near the polyline's points
			for (int i = 0; i < poly.size(); i++)
				{
				int x = poly[i].x(), y = poly[i].y();
					if ( (x >= 0) && (x < image.getCols()) && (y >= 0) && (y < image.getRows()) )
						dest(x, y) = image(x, y);
				}

			setPropertyValue< QVImage<uChar,1> >("Output image", dest);
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv, "Example program for QVision library." );
	
	MyWorker myWorker("worker");
	QVMPlayerCamera camera("Video");
	camera.link(&myWorker, "Input image");

	QVDefaultGUI interface;

	QVImageCanvas imageCanvas("image");
	imageCanvas.linkProperty(myWorker, "Output image");
	imageCanvas.linkROI(&myWorker, "in rect selec");
	imageCanvas.linkPolyline(&myWorker, "in polyline");

	return app.exec();
	}

#endif
	
