/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvpolylinef.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVPOLYLINEF_H
#define QVPOLYLINEF_H

#include <QPointF>

#include <qvipp.h>
#include <QVImage>


class QVPolyline;

/// @class QVPolylineF qvip/qvpolylinef.h
/// @brief Floating point polyline representation.
/// @ingroup qvimagefeatures
///
/// A polyline is a continuous line composed of one or more straight line segments.
/// A QVPolylineF is a implementation of a polyline, specified by a list of the points located at the intersection of the segments.
/// Thus, any QVPolylineF object inherits from the class QList<QPointF>.
///
/// When a polyline is closed, then it is considered to be a polygon. In that case it is supposed to have an implicit final
/// straight line segment, that links the first and the last point in the polyline. The property @ref QVPolylineF::closed will
/// store the value <i>TRUE</i> if the polyline is a polygon, else storing <i>FALSE</i>.
///
/// A QVPolylineF has the property @ref QVPolylineF::direction, that will indicate if it is a direct polyline, or a reverse
/// polyline. It has only meaning for some algorithms, like @ref getBorderContourThreshold.
class QVPolylineF: public QList<QPointF>
	{
	public:
		bool closed, direction;

		QVPolylineF();
		QVPolylineF(const QVPolyline &polyline);
		QVPolylineF(const QVPolylineF &polyline);

		static QVPolylineF ellipse(uInt n, float x, float y, float maxRadio, float minRadio, float ang);
		static QVPolylineF rectangle(float x1, float y1, float x2, float y2);

		operator QVPolyline() const;
	};

/// @brief Simplifies a contour eliminating points of little area using IPE algorithm
/// @ingroup qvimagefeatures
///
/// This function eliminates points of a QVPolylineF, simplificating it.
/// Points are recursively eliminated while a) their distance to the line joining
/// its two immediate neighbours is smaller than a given maximum value, or
/// b) the number of points of the approximated polyline falls below a given
/// number (depending on the chosen method).
///
/// The return value (which indicates the cost of the first not deleted point)
/// and the optional parameter max_removed_cost are useful to estimate
/// a measure of how well the polyline approximation did. A good
/// approximation will have (relative) low max_removed_cost/return_cost ratio,
/// and (absolute) low max_removed_cost value.
///
/// @param polyline polyline to simplify.
/// @param result will store resulting simplified polyline.
/// @param param maximal distance for a point to eliminate it, if
///              maxNumberOfPointsMethod is FALSE, or maximal number of points in the
///              result polyline, if maxNumberOfPointsMethod is TRUE.
/// @param maxNumberOfPointsMethod if TRUE, the procedure ends when the number
///              of points of the resulting polyline is lesser or equal to param.
///              If FALSE, the procedure ends when there are no more points
///              with distance to line given by its neighbours smaller than
///              param.
/// @param intersectLines If TRUE, a post-processing stage adjusts final
///              points using all the deleted points among vertexs to fit
///              straight lines, whose intersection by pairs gives the final
///              points (recommended for greater precision, at a minimal
///              additional cost).
/// @param max_removed_cost If not NULL, pointer to a float value that the
///              procedure will fill with the maximum cost (distance to
///              corresponding line) of all the removed points.
/// @return The function returns the cost value of the first NOT deleted point
///         of the polylyne.
double IterativePointElimination(const QVPolylineF &polyline, QVPolylineF &result,
	const double param, bool maxNumberOfPointsMethod=FALSE,
	bool intersectLines=TRUE, double *max_removed_cost=NULL);

Q_DECLARE_METATYPE(QList<QVPolylineF>);
#endif

