/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvmatrixalgebra.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_linalg.h>

#include <qvmath/qvmatrixalgebra.h>
#include <qvcore/qvdefines.h>

void solveLinear(const QVMatrix &A, QVVector &x, const QVVector &b)
	{
	Q_ASSERT(A.getCols() == A.getRows());
	Q_ASSERT(A.getCols() == x.size());
	Q_ASSERT(A.getCols() == b.size());

	gsl_matrix *gA = A;
	gsl_vector *gB = b;

	gsl_linalg_HH_svx(gA, gB);
	x = gB;

	gsl_matrix_free(gA);
	gsl_vector_free(gB);
	}

void solveLinear(const QVMatrix &A, QVMatrix &X, const QVMatrix &B)
	{
	Q_ASSERT(A.getCols() == A.getRows());
	Q_ASSERT(A.getCols() == X.getRows());
	Q_ASSERT(A.getCols() == B.getRows());

	const int dimN = A.getRows();
	const int numS = X.getCols();
	int signum;

	double *dataX = X.getWriteData();
	const double *dataB = B.getReadData();

	gsl_matrix *a = A;
	gsl_permutation *p = gsl_permutation_alloc(dimN);
	gsl_vector *b = gsl_vector_alloc(dimN);
	gsl_vector *x = gsl_vector_alloc(dimN);

	gsl_linalg_LU_decomp(a, p, &signum);

	for(int sist = 0; sist < numS; sist++)
		{
		for(int i = 0; i < dimN; i++)
			b->data[i] = dataB[i*numS + sist];

		gsl_linalg_LU_solve(a, p, b, x);

		for(int i = 0; i < dimN; i++)
			dataX[i*numS + sist] = x->data[i];
		}

	gsl_matrix_free(a);
	gsl_permutation_free(p);
	gsl_vector_free(b);
	gsl_vector_free(x);
	}

void solveOverDetermined(const QVMatrix &A, QVMatrix &X, const QVMatrix &B)
	{
	Q_ASSERT(A.getCols() <= A.getRows());
	Q_ASSERT(A.getCols() == X.getRows());
	Q_ASSERT(A.getRows() == B.getRows());

	const int dim1 = A.getRows();
	const int dim2 = A.getCols();
	const int numS = X.getCols();

	double *dataX = X.getWriteData();
	const double *dataB = B.getReadData();

	gsl_matrix *u = A;
	gsl_vector *s = gsl_vector_alloc(dim2);
	gsl_matrix *v = gsl_matrix_alloc(dim2, dim2);
	gsl_vector *workV = gsl_vector_alloc(dim2);
	gsl_matrix *workM = gsl_matrix_alloc(dim2,dim2);
	gsl_vector *b = gsl_vector_alloc(dim1);
	gsl_vector *x = gsl_vector_alloc(dim2);

	gsl_linalg_SV_decomp_mod(u, workM, v, s, workV);

	for(int sist = 0; sist < numS; sist++)
		{
		for(int i = 0; i < dim1; i++)
			b->data[i] = dataB[i*numS + sist];

		gsl_linalg_SV_solve(u, v, s, b, x);

		for(int i = 0; i < dim2; i++)
			dataX[i*numS + sist] = x->data[i];
		}

	gsl_matrix_free(u);
	gsl_vector_free(s);
	gsl_matrix_free(v);
	gsl_vector_free(workV);
	gsl_matrix_free(workM);
	gsl_vector_free(b);
	gsl_vector_free(x);
	}

void solveHomogeneousLinear(const QVMatrix &A, QVector<double> &x)
	{
	QVMatrix U, V, S;
	singularValueDecomposition(A, U, V, S);

	x = V.getCol(V.getCols()-1);
	}

void solveHomogeneousLinear2(const QVMatrix &A, QVector<double> &x)
	{
	const int rows = A.getRows(), cols = A.getCols();

	// Initialize A2 matrix and b vector.
	QVMatrix A2(rows, cols-1);

	for (int i = 1; i < cols; i++)
		A2.setCol(i-1, A.getCol(i));

	// Find the best solution
	double minValue = std::numeric_limits<double>::max();
	QVVector bestX(cols);
	for (int i = 0; i <= cols-1; i++)
		{
		QVVector b = A.getCol(i);
		QVMatrix bMat(rows, 1);
		bMat.setCol(0,b);

		QVMatrix x_temp =  bMat.transpose() / A2.transpose();

		if (x_temp.norm2() < minValue)
			{
			bestX = x_temp.getRow(0);
			bestX.insert(i,-1);
			minValue = x_temp.norm2();			
			}

		// Swap actual column
		if (i < cols - 1)
			A2.setCol(i, A.getCol(i));
		}

	x = bestX;
	}

void singularValueDecomposition(const QVMatrix &M, QVMatrix &U, QVMatrix &V, QVMatrix &S)
	{
	const int	dim1 = M.getRows(),
			dim2 = M.getCols();

	gsl_matrix *u = M;
	gsl_vector *s = gsl_vector_alloc (dim2);
	gsl_matrix *v = gsl_matrix_alloc (dim2, dim2);
	gsl_vector *work = gsl_vector_alloc (dim2);

	//gsl_linalg_SV_decomp(u, v, s, work);

	gsl_matrix *X = gsl_matrix_alloc (dim2,dim2);
	gsl_linalg_SV_decomp_mod(u, X, v, s, work);
	gsl_matrix_free(X);

	/// @todo borrar el código comentado, cuando se compruebe que el nuevo funciona bien.

	// Esto:
	U = u;
	V = v;
	S = QVMatrix::diagonal(s);

	// ... es equivalente a esto:
	/*U = QVMatrix(dim1, dim2);
	V = QVMatrix(dim2, dim2);
	S = QVMatrix(dim2, dim2);

	double	*dataU = U.getWriteData(),
		*dataV = V.getWriteData(),
		*dataS = S.getWriteData();

	for (int i = 0; i < dim1; i++)
		for (int j = 0; j < dim2; j++)
			dataU[i*dim2 + j] = u->data[i*dim2+j];

	for (int i = 0; i < dim2; i++)
		{
		for (int j = 0; j < dim2; j++)
			{
			dataV[i*dim2 + j] = v->data[i*dim2+j];
			dataS[i*dim2 + j] = 0;
			}
		dataS[i*dim2+i] = s->data[i];
		}*/

	gsl_matrix_free(u);
	gsl_vector_free(s);
	gsl_matrix_free(v);
	gsl_vector_free(work);
	}

void LUDecomposition(const QVMatrix &M, QVMatrix &L, QVMatrix &U, QVMatrix &P)
	{
	const int	dim1 = M.getRows(),
			dim2 = M.getCols();

	gsl_matrix *a = M;
	gsl_permutation *p = gsl_permutation_alloc(dim2);

	int signum;


	gsl_linalg_LU_decomp(a, p, &signum);


	L = QVMatrix(dim1, dim2);
	U = QVMatrix(dim1, dim2);
	P = QVMatrix(dim1, dim2);

	double	*dataL = L.getWriteData(),
		*dataU = U.getWriteData(),
		*dataP = P.getWriteData();

	for (int i = 0; i < dim1; i++)
		for (int j = 0; j < dim2; j++)
			{
			if (j > i)
				{
				dataU[i*dim2 + j] = a->data[i*dim2+j];
				dataL[i*dim2 + j] = 0;
				}
			else if (j < i)
				{
				dataU[i*dim2 + j] = 0;
				dataL[i*dim2 + j] = a->data[i*dim2+j];
				}
			else
				{
				dataU[i*dim2 + j] = a->data[i*dim2+j];
				dataL[i*dim2 + j] = 1;
				}
			}

	for (int i = 0; i < dim1; i++)
		for (int j = 0; j < dim2; j++)
			dataP[i*dim2 + j] = 0;

	for (int j = 0; j < dim2; j++)
		dataP[(p->data[j])*dim2 + j] = 1;


	gsl_matrix_free(a);
	gsl_permutation_free(p);
	}

void CholeskyDecomposition(const QVMatrix &M, QVMatrix &L)
	{
	const int	dim1 = M.getRows(),
			dim2 = M.getCols();

	gsl_matrix *a = M;


	gsl_linalg_cholesky_decomp(a);


	L = QVMatrix(dim1, dim2);
	double	*dataL = L.getWriteData();

	for (int i = 0; i < dim1; i++)
		for (int j = 0; j < dim2; j++)
			{
			if (j <= i)
				dataL[i*dim2 + j] = a->data[i*dim2+j];
			else
				dataL[i*dim2 + j] = 0;
			}

	gsl_matrix_free(a);
	}

void QRDecomposition(const QVMatrix &M, QVMatrix &Q, QVMatrix &R)
	{
	const int	dim1 = M.getRows(),
			dim2 = M.getCols(),
			min = (dim1<dim2 ? dim1: dim2);

	gsl_matrix *a = M;
	gsl_matrix *q = gsl_matrix_alloc(dim1, dim1);
	gsl_matrix *r = gsl_matrix_alloc(dim1, dim2);
	gsl_vector *tau = gsl_vector_alloc(min);

	gsl_linalg_QR_decomp(a, tau);
	gsl_linalg_QR_unpack (a, tau, q, r);

	Q = QVMatrix(dim1, dim2);
	R = QVMatrix(dim1, dim2);

	double	*dataQ = Q.getWriteData(),
		*dataR = R.getWriteData();

	for (int i = 0; i < dim1; i++)
		for (int j = 0; j < dim1; j++)
			dataQ[i*dim2 + j] = q->data[i*dim2+j];

	for (int i = 0; i < dim1; i++)
		for (int j = 0; j < dim2; j++)
			dataR[i*dim2 + j] = r->data[i*dim2+j];

	gsl_matrix_free(a);
	gsl_matrix_free(q);
	gsl_matrix_free(r);
	gsl_vector_free(tau);
	}

QVMatrix pseudoInverse(const QVMatrix &M)
	{
	if (M.getRows() < M.getCols())
		return pseudoInverse(M.transpose()).transpose();

	QVMatrix U, V, S;

	singularValueDecomposition(M, U, V, S);

	const int dim2 = M.getCols();

	double *dataBufferS = S.getWriteData();
	for (int i = 0; i < dim2; i++)
		dataBufferS[i*dim2+i] = 1/dataBufferS[i*dim2+i];

	return V * S * U.transpose();
 	}

double determinant(const QVMatrix &M)
	{
	Q_ASSERT(M.getRows() == M.getCols());
	Q_ASSERT(M.getRows() > 1);

	QVMatrix U, V, S;

	/// @TODO Por ahora con la SVD, quizá sería mejor de otro modo...
	singularValueDecomposition(M, U, V, S);

	double value=1.0;
	const int dim = M.getRows();

	double *dataBufferS = S.getWriteData();
	for (int i = 0; i < dim; i++)
		value *= dataBufferS[i*dim+i];

	return value;
 	}

double BhattacharyyaDistance(const QVVector &m1, const QVMatrix &S1, const QVVector &m2, const QVMatrix &S2)
	{
	QVVector diff = m2-m1;
	QVMatrix mS = (S1+S2)/2;
	QVMatrix inv = mS.inverse();
	double value = (diff*(inv*diff))/8 + log(mS.det()/sqrt(S1.det()*S2.det()))/2;
	return value;
	}

void eigenDecomposition(const QVMatrix &M, QVVector &eigVals, QVMatrix &eigVecs)
	{
	// Check matrix M is symetric
	Q_ASSERT(M.getCols() == M.getRows());

	double data[M.getDataSize()];

	const double *dataBuffer = M.getReadData();
	for(int i = 0; i < M.getDataSize(); i++)
		data[i] = dataBuffer[i];
	
	const int dim = M.getRows();
	gsl_matrix_view m = gsl_matrix_view_array (data, dim, dim);
	gsl_vector *eval = gsl_vector_alloc (dim);
	gsl_matrix *evec = gsl_matrix_alloc (dim, dim);

	gsl_eigen_symmv_workspace * w = gsl_eigen_symmv_alloc (dim);
	gsl_eigen_symmv (&m.matrix, eval, evec, w);
	gsl_eigen_symmv_free (w);
	gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_DESC);

	/// @todo borrar el código comentado, cuando se compruebe que el nuevo funciona bien.

	// Esto:
	eigVals = eval;
	eigVecs = evec;
	eigVecs = eigVecs.transpose();

	// ... es equivalente a esto:
	/*
	eigVals = QVector<double>(dim);
	eigVecs = QVMatrix(dim, dim);

	double	*dataVecs = eigVecs.getWriteData();

	for (int i = 0; i < dim; i++)
		{
		eigVals[i] = eval->data[i];
		for (int j = 0; j < dim; j++)
			dataVecs[i*dim + j] = evec->data[j*dim + i];
		}
	*/

	gsl_vector_free (eval);
	gsl_matrix_free (evec);
	}

// Adjust a line to given first and second order moments, and returns ratio of
// eigenvalues:

// Adjust a line to given first and second order moments, and returns ratio of 
// eigenvalues: 
double homogLineFromMoments(double x,double y,double xx,double xy,double yy,double &a,double &b,double &c)
        {
	double a11=xx-x*x, a12=xy-x*y, a22=yy-y*y, temp, e1, e2, angle, cosangle, sinangle;

	// Validity check:
	temp = sqrt(a11*a11+4*a12*a12-2*a11*a22+a22*a22);
	e1 = a11+a22-temp;
	e2 = a11+a22+temp;
	if(e2<EPSILON)
		{
		/*std::cerr << " a11=" << a11 << " a12=" << a12 << " a22=" << a22  
		<< " x=" << x << " y=" << y << " xx=" << xx << " xy=" << xy << " yy=" << yy  
		<< ": Imposible moments in homogLineFromMoments!\n";*/ 
		return 1.0; 
		} 
	if(fabs(e1)/e2 > 0.9)
		{
		/*std::cerr << "Too high ratio of eigenvalues e1=" << e1 << "e2=" << e2 <<": No principal direction in homogLineFromMoments!\n"; 
		std::cerr << " a11=" << a11 << " a12=" << a12 << " a22=" << a22 << "\n";*/ 
		return fabs(e1)/e2;
		} 

	if(fabs(a12)>EPSILON)
		angle = atan2(2*a12,a11-a22+temp);
	else
	if(a11>=a22)
		angle = 0;
	else
		angle = PI/2;
	if(angle < 0)
		angle += PI;
	cosangle = cos(angle); sinangle = sin(angle);

	//Standard deviation in perpendicular direction:
	//desv_perp = sqrt(fabs(a11+a22-temp)/2.0);
	a = -sinangle; b = cosangle; c = x*sinangle-y*cosangle;
	return fabs(e1)/e2;
	}

QVVector regressionLine(const QVMatrix &points)
	{
	/// @todo esto no funciona.
	double x = 0, y = 0, xx = 0, yy = 0, xy = 0;
	const int rows = points.getRows();

	for (int i = 0; i < rows; i++)
		{
		double xActual = points(i,0), yActual = points(i,1);
		x += xActual;
		y += yActual;
		xx += xActual*xActual;
		xy += xActual*yActual;
		yy += yActual*yActual;
		}

	x /= rows; y /= rows; xx /= rows; xy /= rows; yy /= rows;

	double a, b, c;
	if (homogLineFromMoments(x,y,xx,xy,yy,a,b,c))
		{		
		QVVector result(3);
		result[0] = a; result[1] = b; result[2] = c;
		return result;
		}
	else
		return QVVector();
	}
