/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvvector.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <QString>
#include <QVVector>
#include <QVMatrix>

QVVector::QVVector(const QVMatrix &matrix): QVector<double>(matrix.getCols() * matrix.getRows())
	{
	const int n = size();
	const double *matrixData = matrix.getReadData();

	for(int i = 0; i < n; i++)
		operator[](i) = matrixData[i];
	}

double QVVector::dotProduct(const QVVector &vector) const
	{
	Q_ASSERT(size() == vector.size());

	double accum = 0;
	for (int i = 0; i < size(); i++)
		accum += at(i) * vector[i];

	return accum;
	}

QVVector QVVector::crossProduct(const QVVector &vector) const
	{
	Q_ASSERT(size() == vector.size());
	Q_ASSERT(size() == 3);

	const double	x1 = at(0), y1 = at(1), z1 = at(2),
			x2 = vector[0], y2 = vector[1], z2 = vector[2];

	QVVector v(3);

	v[0] = -y2*z1 + y1*z2;
	v[1] = x2*z1 - x1*z2;
	v[2] = -x2*y1 + x1*y2;

	return v;
	}

QVVector QVVector::operator*(const QVMatrix &matrix) const
	{
	Q_ASSERT(size() == matrix.getRows());
	return QVMatrix(*this).dotProduct(matrix).getRow(0);
	}

QVVector QVVector::add(const QVVector &vector) const
	{
	Q_ASSERT(size() == vector.size());

	QVVector result(size());
	for (int i = 0; i < size(); i++)
		result[i] = at(i) + vector[i];

	return result;
	}

QVVector QVVector::substract(const QVVector &vector) const
	{
	Q_ASSERT(size() == vector.size());

	QVVector result(size());
	for (int i = 0; i < size(); i++)
		result[i] = at(i) - vector[i];

	return result;
	}

bool QVVector::equals(const QVVector &vector) const
	{
	if (size() != vector.size())
		return false;

	for (int i = 0; i < size(); i++)
		if (at(i) != vector[i])
			return false;

	return true;
	}

//////////////////////////////////////////////

std::ostream& operator << ( std::ostream &os, const QVVector &vector )
	{
	const int size = vector.size();

	os << "QVVector (" << size << ") [ ";

	for (int i = 0; i < size; i++)
		os << qPrintable(QString("%1").arg(vector[i], -8, 'f', 6)) << " ";

	os << "]";
	return os;
	}

uint qHash(const QVVector &vector)
	{
	const int size = vector.size();
	double accum = 0;
	for (int i = 0; i < size; i++)
		accum += vector[i] / vector[size-i-1];

	return (uint) ((100000 * accum) / ((double) size));
	}

